@@ -30,18 +30,6 @@ pub struct InstData {
 }
 
 impl InstData {
-    pub fn op(def: usize, uses: &[usize]) -> InstData {
-        let mut operands = vec![Operand::reg_def(VReg::new(def, RegClass::Int))];
-        for &u in uses {
-            operands.push(Operand::reg_use(VReg::new(u, RegClass::Int)));
-        }
-        InstData {
-            op: InstOpcode::Op,
-            operands,
-            clobbers: vec![],
-            is_safepoint: false,
-        }
-    }
     pub fn branch() -> InstData {
         InstData {
             op: InstOpcode::Branch,
@@ -145,8 +133,10 @@ impl Function for Func {
 
     fn spillslot_size(&self, regclass: RegClass) -> usize {
         match regclass {
+            // Test the case where 2 classes share the same
             RegClass::Int => 1,
-            RegClass::Float | RegClass::Vector => 2,
+            RegClass::Float => 1,
+            RegClass::Vector => 2,
         }
     }
 }
@@ -182,7 +172,7 @@ impl FuncBuilder {
         let b = Block::new(self.f.blocks.len());
         self.f
             .blocks
-            .push(InstRange::forward(Inst::new(0), Inst::new(0)));
+            .push(InstRange::new(Inst::new(0), Inst::new(0)));
         self.f.block_preds.push(vec![]);
         self.f.block_succs.push(vec![]);
         self.f.block_params_in.push(vec![]);
@@ -227,13 +217,19 @@ impl FuncBuilder {
                 self.f.insts.push(inst.clone());
             }
             let end_inst = self.f.insts.len();
-            *blockrange = InstRange::forward(Inst::new(begin_inst), Inst::new(end_inst));
+            *blockrange = InstRange::new(Inst::new(begin_inst), Inst::new(end_inst));
         }
 
         self.f
     }
 }
 
+impl Arbitrary<'_> for RegClass {
+    fn arbitrary(u: &mut Unstructured) -> ArbitraryResult<Self> {
+        Ok(*u.choose(&[RegClass::Int, RegClass::Float, RegClass::Vector])?)
+    }
+}
+
 impl Arbitrary<'_> for OperandConstraint {
     fn arbitrary(u: &mut Unstructured) -> ArbitraryResult<Self> {
         Ok(*u.choose(&[OperandConstraint::Any, OperandConstraint::Reg])?)
@@ -271,10 +267,6 @@ pub struct Options {
     pub fixed_regs: bool,
     pub fixed_nonallocatable: bool,
     pub clobbers: bool,
-    pub control_flow: bool,
-    pub reducible: bool,
-    pub block_params: bool,
-    pub always_local_uses: bool,
     pub reftypes: bool,
 }
 
@@ -285,10 +277,6 @@ impl core::default::Default for Options {
             fixed_regs: false,
             fixed_nonallocatable: false,
             clobbers: false,
-            control_flow: true,
-            reducible: false,
-            block_params: true,
-            always_local_uses: false,
             reftypes: false,
         }
     }
@@ -326,12 +314,9 @@ impl Func {
         let mut from = 0;
         let mut out_blocks = vec![];
         let mut in_blocks = vec![];
-        // For reducibility, if selected: enforce strict nesting of backedges
-        let mut max_backedge_src = 0;
-        let mut min_backedge_dest = num_blocks;
         while from < num_blocks {
             in_blocks.push(from);
-            if num_blocks > 3 && from < num_blocks - 3 && bool::arbitrary(u)? && opts.control_flow {
+            if num_blocks > 3 && from < num_blocks - 3 && bool::arbitrary(u)? {
                 // To avoid critical edges, we use from+1 as an edge
                 // block, and advance `from` an extra block; `from+2`
                 // will be the next normal iteration.
@@ -346,18 +331,7 @@ impl Func {
             from += 1;
         }
         for pred in out_blocks {
-            let mut succ = *u.choose(&in_blocks[..])?;
-            if opts.reducible && (pred >= succ) {
-                if pred < max_backedge_src || succ > min_backedge_dest {
-                    // If the chosen edge would result in an
-                    // irreducible CFG, just make this a diamond
-                    // instead.
-                    succ = pred + 2;
-                } else {
-                    max_backedge_src = pred;
-                    min_backedge_dest = succ;
-                }
-            }
+            let succ = *u.choose(&in_blocks[..])?;
             builder.add_edge(Block::new(pred), Block::new(succ));
         }
 
@@ -380,7 +354,7 @@ impl Func {
         for block in 0..num_blocks {
             let mut vregs = vec![];
             for _ in 0..u.int_in_range(5..=15)? {
-                let vreg = VReg::new(builder.f.num_vregs, RegClass::Int);
+                let vreg = VReg::new(builder.f.num_vregs, RegClass::arbitrary(u)?);
                 builder.f.num_vregs += 1;
                 vregs.push(vreg);
                 if opts.reftypes && bool::arbitrary(u)? {
@@ -406,17 +380,18 @@ impl Func {
                 }
             }
             vregs_by_block.push(vregs.clone());
-            vregs_by_block_to_be_defined.push(vec![]);
+            let mut vregs_to_be_defined = vec![];
             let mut max_block_params = u.int_in_range(0..=core::cmp::min(3, vregs.len() / 3))?;
             for &vreg in &vregs {
-                if block > 0 && opts.block_params && bool::arbitrary(u)? && max_block_params > 0 {
+                if block > 0 && bool::arbitrary(u)? && max_block_params > 0 {
                     block_params[block].push(vreg);
                     max_block_params -= 1;
                 } else {
-                    vregs_by_block_to_be_defined.last_mut().unwrap().push(vreg);
+                    vregs_to_be_defined.push(vreg);
                 }
             }
-            vregs_by_block_to_be_defined.last_mut().unwrap().reverse();
+            vregs_to_be_defined.reverse();
+            vregs_by_block_to_be_defined.push(vregs_to_be_defined);
             builder.set_block_params_in(Block::new(block), &block_params[block][..]);
         }
 
@@ -439,12 +414,10 @@ impl Func {
                 let mut allocations = vec![Allocation::none()];
                 for _ in 0..u.int_in_range(0..=3)? {
                     let vreg = if avail.len() > 0
-                        && (opts.always_local_uses
-                            || remaining_nonlocal_uses == 0
-                            || bool::arbitrary(u)?)
+                        && (remaining_nonlocal_uses == 0 || bool::arbitrary(u)?)
                     {
                         *u.choose(&avail[..])?
-                    } else if !opts.always_local_uses {
+                    } else {
                         let def_block = choose_dominating_block(
                             &builder.idom[..],
                             Block::new(block),
@@ -458,8 +431,6 @@ impl Func {
                         }
                         remaining_nonlocal_uses -= 1;
                         *u.choose(&vregs_by_block[def_block.index()])?
-                    } else {
-                        break;
                     };
                     let use_constraint = OperandConstraint::arbitrary(u)?;
                     operands.push(Operand::new(
@@ -476,20 +447,22 @@ impl Func {
                     let op = operands[0];
                     debug_assert_eq!(op.kind(), OperandKind::Def);
                     let reused = u.int_in_range(1..=(operands.len() - 1))?;
-                    operands[0] = Operand::new(
-                        op.vreg(),
-                        OperandConstraint::Reuse(reused),
-                        op.kind(),
-                        OperandPos::Late,
-                    );
-                    // Make sure reused input is a Reg.
-                    let op = operands[reused];
-                    operands[reused] = Operand::new(
-                        op.vreg(),
-                        OperandConstraint::Reg,
-                        op.kind(),
-                        OperandPos::Early,
-                    );
+                    if op.class() == operands[reused].class() {
+                        operands[0] = Operand::new(
+                            op.vreg(),
+                            OperandConstraint::Reuse(reused),
+                            op.kind(),
+                            OperandPos::Late,
+                        );
+                        // Make sure reused input is a Reg.
+                        let op = operands[reused];
+                        operands[reused] = Operand::new(
+                            op.vreg(),
+                            OperandConstraint::Reg,
+                            op.kind(),
+                            OperandPos::Early,
+                        );
+                    }
                 } else if opts.fixed_regs && bool::arbitrary(u)? {
                     let mut fixed_early = vec![];
                     let mut fixed_late = vec![];
@@ -497,7 +470,21 @@ impl Func {
                         // Pick an operand and make it a fixed reg.
                         let i = u.int_in_range(0..=(operands.len() - 1))?;
                         let op = operands[i];
-                        let fixed_reg = PReg::new(u.int_in_range(0..=62)?, RegClass::Int);
+                        let fixed_reg = PReg::new(u.int_in_range(0..=62)?, op.class());
+                        if op.kind() == OperandKind::Def && op.pos() == OperandPos::Early {
+                            // Early-defs with fixed constraints conflict with
+                            // any other fixed uses of the same preg.
+                            if fixed_late.contains(&fixed_reg) {
+                                break;
+                            }
+                        }
+                        if op.kind() == OperandKind::Use && op.pos() == OperandPos::Late {
+                            // Late-use with fixed constraints conflict with
+                            // any other fixed uses of the same preg.
+                            if fixed_early.contains(&fixed_reg) {
+                                break;
+                            }
+                        }
                         let fixed_list = match op.pos() {
                             OperandPos::Early => &mut fixed_early,
                             OperandPos::Late => &mut fixed_late,
@@ -505,22 +492,6 @@ impl Func {
                         if fixed_list.contains(&fixed_reg) {
                             break;
                         }
-                        if op.kind() != OperandKind::Def && op.pos() == OperandPos::Late {
-                            // Late-uses/mods with fixed constraints
-                            // can't be allowed if we're allowing
-                            // different constraints at Early and
-                            // Late, because we can't move something
-                            // into a location between Early and
-                            // Late. Differing constraints only make
-                            // sense if the instruction itself
-                            // produces the newly-constrained values.
-                            break;
-                        }
-                        if op.kind() != OperandKind::Use && op.pos() == OperandPos::Early {
-                            // Likewise, we can *only* allow uses for
-                            // fixed constraints at Early.
-                            break;
-                        }
                         fixed_list.push(fixed_reg);
                         operands[i] = Operand::new(
                             op.vreg(),
@@ -535,10 +506,13 @@ impl Func {
                         if clobbers.iter().any(|r| r.hw_enc() == reg) {
                             break;
                         }
-                        clobbers.push(PReg::new(reg, RegClass::Int));
+                        clobbers.push(PReg::new(reg, RegClass::arbitrary(u)?));
                     }
                 } else if opts.fixed_nonallocatable && bool::arbitrary(u)? {
-                    operands.push(Operand::fixed_nonallocatable(PReg::new(63, RegClass::Int)));
+                    operands.push(Operand::fixed_nonallocatable(PReg::new(
+                        63,
+                        RegClass::arbitrary(u)?,
+                    )));
                 }
 
                 let is_safepoint = opts.reftypes
@@ -565,18 +539,30 @@ impl Func {
                 let mut params = vec![];
                 for &succ in &builder.f.block_succs[block] {
                     let mut args = vec![];
-                    for _ in 0..builder.f.block_params_in[succ.index()].len() {
+                    for i in 0..builder.f.block_params_in[succ.index()].len() {
                         let dom_block = choose_dominating_block(
                             &builder.idom[..],
                             Block::new(block),
                             false,
                             u,
                         )?;
-                        let vreg = if dom_block.is_valid() && bool::arbitrary(u)? {
-                            u.choose(&vregs_by_block[dom_block.index()][..])?
+
+                        // Look for a vreg with a suitable class. If no
+                        // suitable vreg is available then we error out, which
+                        // causes the fuzzer to skip this function.
+                        let vregs = if dom_block.is_valid() && bool::arbitrary(u)? {
+                            &vregs_by_block[dom_block.index()][..]
                         } else {
-                            u.choose(&avail[..])?
+                            &avail[..]
                         };
+                        let suitable_vregs: Vec<_> = vregs
+                            .iter()
+                            .filter(|vreg| {
+                                vreg.class() == builder.f.block_params_in[succ.index()][i].class()
+                            })
+                            .copied()
+                            .collect();
+                        let vreg = u.choose(&suitable_vregs)?;
                         args.push(*vreg);
                     }
                     params.push(args);
@@ -656,13 +642,29 @@ impl core::fmt::Debug for Func {
 }
 
 pub fn machine_env() -> MachineEnv {
-    fn regs(r: core::ops::Range<usize>) -> Vec<PReg> {
-        r.map(|i| PReg::new(i, RegClass::Int)).collect()
+    fn regs(r: core::ops::Range<usize>, c: RegClass) -> Vec<PReg> {
+        r.map(|i| PReg::new(i, c)).collect()
     }
-    let preferred_regs_by_class: [Vec<PReg>; 3] = [regs(0..24), vec![], vec![]];
-    let non_preferred_regs_by_class: [Vec<PReg>; 3] = [regs(24..32), vec![], vec![]];
+    let preferred_regs_by_class: [Vec<PReg>; 3] = [
+        regs(0..24, RegClass::Int),
+        regs(0..24, RegClass::Float),
+        regs(0..24, RegClass::Vector),
+    ];
+    let non_preferred_regs_by_class: [Vec<PReg>; 3] = [
+        regs(24..32, RegClass::Int),
+        regs(24..32, RegClass::Float),
+        regs(24..32, RegClass::Vector),
+    ];
     let scratch_by_class: [Option<PReg>; 3] = [None, None, None];
-    let fixed_stack_slots = regs(32..63);
+    let fixed_stack_slots = (32..63)
+        .flat_map(|i| {
+            [
+                PReg::new(i, RegClass::Int),
+                PReg::new(i, RegClass::Float),
+                PReg::new(i, RegClass::Vector),
+            ]
+        })
+        .collect();
     // Register 63 is reserved for use as a fixed non-allocatable register.
     MachineEnv {
         preferred_regs_by_class,
